;; Generated package description from gcmh.el  -*- no-byte-compile: t -*-
(define-package "gcmh" "0.2.1.0.20201116.225142" "the Garbage Collector Magic Hack" '((emacs "24")) :commit "0089f9c3a6d4e9a310d0791cf6fa8f35642ecfd9" :authors '(("Andrea Corallo" . "akrl@sdf.org")) :maintainer '(nil . "akrl@sdf.org") :keywords '("internal") :url "https://gitlab.com/koral/gcmh")
